@@ -16,7 +16,6 @@ import (
 	"log"
 	"net"
 	"net/http"
-	_ "net/http/pprof"
 	"net/url"
 	"os"
 	"os/signal"
@@ -664,8 +663,14 @@ func syncthingMain(runtimeOptions RuntimeOptions) {
 		}
 	}
 
+	if cfg.Raw().OriginalVersion == 15 {
+		// The config version 15->16 migration is about handling ignores and
+		// delta indexes and requires that we drop existing indexes that
+		// have been incorrectly ignore filtered.
+		ldb.DropDeltaIndexIDs()
+	}
+
 	m := model.NewModel(cfg, myID, myDeviceName(cfg), "syncthing", Version, ldb, protectedFiles)
-	cfg.Subscribe(m)
 
 	if t := os.Getenv("STDEADLOCKTIMEOUT"); len(t) > 0 {
 		it, err := strconv.Atoi(t)
@@ -1122,6 +1127,8 @@ func cleanConfigDirectory() {
 		"panic-*.log":      7 * 24 * time.Hour,  // keep panic logs for a week
 		"audit-*.log":      7 * 24 * time.Hour,  // keep audit logs for a week
 		"index":            14 * 24 * time.Hour, // keep old index format for two weeks
+		"index-v0.11.0.db": 14 * 24 * time.Hour, // keep old index format for two weeks
+		"index-v0.13.0.db": 14 * 24 * time.Hour, // keep old index format for two weeks
 		"index*.converted": 14 * 24 * time.Hour, // keep old converted indexes for two weeks
 		"config.xml.v*":    30 * 24 * time.Hour, // old config versions for a month
 		"*.idx.gz":         30 * 24 * time.Hour, // these should for sure no longer exist
